using System;
using System.Collections;
using System.Reflection;
using System.Runtime.InteropServices;
using PorygonSharp.UserData;

namespace PorygonSharp.ScriptType
{
    internal static class ScriptTypeHandler
    {
        internal static IntPtr? GetScriptType(Type t)
        {
            if (t == null)
            {
                return CreateScriptType(TypeClass.Nil);
            }
            if (t.IsEnum && !t.IsGenericParameter)
            {
                if (UserDataHandler.IsTypeRegistered(t))
                    return UserDataHandler.CreateUserDataType(t);
                return CreateNumericScriptType(true, false);
            }
            var typeCode = Type.GetTypeCode(t);
            switch (typeCode)
            {
                case TypeCode.Boolean:
                    return CreateScriptType(TypeClass.Bool);
                case TypeCode.Byte:
                case TypeCode.SByte:
                case TypeCode.Int16:
                case TypeCode.Int32:
                case TypeCode.Int64:
                case TypeCode.UInt16:
                case TypeCode.UInt32:
                case TypeCode.UInt64:
                    return CreateNumericScriptType(true, false);
                case TypeCode.Decimal:
                case TypeCode.Double:
                case TypeCode.Single:
                    return CreateNumericScriptType(true, true);
                case TypeCode.Char:
                case TypeCode.String:
                    return CreateStringScriptType(false, 0);
                case TypeCode.Object:
                    if (t == typeof(void))
                    {
                        return CreateScriptType(TypeClass.Nil);
                    }
                    if (t == typeof(object))
                    {
                        return CreateScriptType(TypeClass.Any);
                    }
                    if (UserDataHandler.IsTypeRegistered(t))
                    {
                        return UserDataHandler.CreateUserDataType(t);
                    }
                    if (typeof(IList).IsAssignableFrom(t))
                    {
                        return CreateUserDataListType(t);
                    }
                    if (typeof(IDictionary).IsAssignableFrom(t))
                    {
                        return CreateUserDataDictionaryType(t);
                    }

                    if (typeof(Delegate).IsAssignableFrom(t))
                    {
                        return GetFunctionScriptType(t.GetMethod("Invoke"));
                    }

                    var attr = t.GetCustomAttribute<Attributes.PorygonUserdataAttribute>();
                    if (attr != null)
                    {
                        UserDataHandler.RegisterType(attr.Identifier, t);
                        return UserDataHandler.CreateUserDataType(t);
                    }
                    return null;
                default:
                    return null;
            }
        }

        internal static IntPtr? GetFunctionScriptType(MethodInfo info)
        {
            try
            {
                var returnType     = GetScriptType(info.ReturnType);
                if (!returnType.HasValue)
                    return null;
                var parameters     = info.GetParameters();
                var parameterFuncs = new IntPtr[parameters.Length];
                for (var index = 0; index < parameters.Length; index++)
                {
                    var parameter = parameters[index];
                    var parameterType = GetScriptType(parameter.ParameterType);
                    if (!parameterType.HasValue)
                    {
                        return null;
                    }
                    parameterFuncs[index] = parameterType.Value;
                }

                return CreateUserDataFunctionScriptType(returnType.Value, parameterFuncs, parameters.Length);
            }
            catch
            {
                return IntPtr.Zero;
            }
        }

        private static IntPtr? CreateUserDataListType(Type t)
        {
            var keyType = CreateNumericScriptType(true, false);
            var valType = t.IsArray ? t.GetElementType() : t.GenericTypeArguments[0];
            var valueType = GetScriptType(valType);
            if (valueType.HasValue)
                return CreateCollectionType(keyType, valueType.Value);
            return null;
        }
        
        private static IntPtr? CreateUserDataDictionaryType(Type t)
        {
            var keyType = GetScriptType(t.GenericTypeArguments[0]);
            var valueType = GetScriptType(t.GenericTypeArguments[1]);
            if (keyType.HasValue && valueType.HasValue)
                return CreateCollectionType(keyType.Value, valueType.Value);
            return null;
        }
        
        [DllImport("PorygonLang", EntryPoint = "CreateScriptType", CallingConvention = CallingConvention.Cdecl)]
        private static extern IntPtr CreateScriptType(TypeClass t);

        [DllImport("PorygonLang", EntryPoint = "CreateNumericScriptType", CallingConvention = CallingConvention.Cdecl)]
        internal static extern IntPtr CreateNumericScriptType(bool isAware, bool isFloat);

        [DllImport("PorygonLang", EntryPoint = "CreateStringScriptType", CallingConvention = CallingConvention.Cdecl)]
        private static extern IntPtr CreateStringScriptType(bool knownAtBind, uint hash);

        [DllImport("PorygonLang", EntryPoint = "CreateUserDataFunctionScriptType", CallingConvention = CallingConvention.Cdecl)]
        private static extern IntPtr CreateUserDataFunctionScriptType(IntPtr returnType, IntPtr[] parameters, int parameterCount);
        
        [DllImport("PorygonLang", EntryPoint = "CreateCollectionType", CallingConvention = CallingConvention.Cdecl)]
        private static extern IntPtr CreateCollectionType(IntPtr keyType, IntPtr valueType);
    }
}